#!/usr/bin/python3
# -*- coding: UTF-8 -*-

import sys, os.path
import sqlite3
import string, base64

from .color import Color

class Database(object):
    dbName = "stats.db"

    def __init__(self, auto_create = True, db_name=None):

        if db_name is None:
            self.dbName = "stats.db"
        else:
            self.dbName = db_name

        if not os.path.isfile(self.dbName):
            if auto_create:
                self.createDB()
            else:
                raise Exception("Database not found")

    def checkOpen(self):
        conn = sqlite3.connect(self.dbName)
        cursor = conn.cursor()

        cursor.execute("""
        select count(*) as cnt from [stats];
        """)

        conn.close()


    def insertUri(self, uri, result_code, result_length):
        try:
            conn = sqlite3.connect(self.dbName)
            cursor = conn.cursor()

            cursor.execute("""
            insert into [stats] ([uri], [result_code], [result_length])
            VALUES (?,?,?);
            """, (uri, result_code, result_length,))

            conn.commit()

            conn.close()

        except Exception as e:
            Color.pl('{!} {R}Error inserting data:{O} %s{W}' % str(e))
        pass

    def insertStatsL1(self, hash, path):
        try:
            conn = sqlite3.connect(self.dbName)
            cursor = conn.cursor()

            cursor.execute("""
            insert or ignore into [summarized_l1] ([hash], [path])
            VALUES (?,?);
            """, (hash, path,))

            conn.commit()

            conn.close()

        except Exception as e:
            Color.pl('{!} {R}Error inserting data:{O} %s{W}' % str(e))
        pass

    def clearSummarized(self):
        conn = sqlite3.connect(self.dbName)

        cursor = conn.cursor()
        cursor.execute("""
        delete from [summarized_l1];
        """)

        cursor = conn.cursor()
        cursor.execute("""
        delete from [summarized_l2];
        """)

        conn.commit()

        conn.close()

    def insertStatsL2(self, word):
        try:

            if word is None or word.strip() == "":
                return;

            conn = sqlite3.connect(self.dbName)
            cursor = conn.cursor()

            cursor.execute("""
            insert or ignore into [summarized_l2] ([word])
            VALUES (?);
            """, (word,))

            cursor = conn.cursor()
            cursor.execute("""
            update [summarized_l2] set hits = (hits + 1) where [word] = ?;
            """, (word,))

            conn.commit()

            conn.close()

        except Exception as e:
            Color.pl('{!} {R}Error inserting data:{O} %s{W}' % str(e))
        pass

    def selectStats(self):

        ret = []

        conn = sqlite3.connect(self.dbName)
        cursor = conn.cursor()

        cursor.execute("""
        select distinct uri from [stats] where result_code = 200;
        """)

        data = cursor.fetchall()
        if data:
            for row in data:
                ret.append(row[0])

        conn.close()

        return ret

    def selectStatsL1(self):

        ret = []

        conn = sqlite3.connect(self.dbName)
        cursor = conn.cursor()

        cursor.execute("""
        select path from [summarized_l1];
        """)

        data = cursor.fetchall()
        if data:
            for row in data:
                ret.append(row[0])

        conn.close()

        return ret

    def selectStatsL2(self):

        ret = []

        conn = sqlite3.connect(self.dbName)
        cursor = conn.cursor()

        cursor.execute("""
        select word, hits from [summarized_l2] order by hits desc;
        """)

        data = cursor.fetchall()
        if data:
            for row in data:
                ret.append((row[0], row[1]))

        conn.close()

        return ret

    def createDB(self):
        # conectando...
        conn = sqlite3.connect(self.dbName)
        # definindo um cursor
        cursor = conn.cursor()

        # criando a tabela (schema)
        cursor.execute("""
            CREATE TABLE IF NOT EXISTS [stats] (
                id_uri INTEGER PRIMARY KEY AUTOINCREMENT,
                uri TEXT NOT NULL, 
                result_code INTEGER NOT NULL DEFAULT(0),
                result_length INTEGER NOT NULL DEFAULT(0),
                created_date datetime not null DEFAULT (datetime('now','localtime'))
            );
        """)

        cursor.execute("""
            CREATE TABLE IF NOT EXISTS [summarized_l1] (
                hash TEXT NOT NULL,
                path TEXT NOT NULL,
                PRIMARY KEY (hash, path)
            );
        """)

        cursor.execute("""
            CREATE TABLE IF NOT EXISTS [summarized_l2] (
                word TEXT NOT NULL PRIMARY KEY,
                hits INTEGER NOT NULL DEFAULT(0)
            );
        """)

        #print('DB criado com sucesso.')
        # desconectando...
        conn.close()



#!/usr/bin/python3
# -*- coding: UTF-8 -*-

try:
    from .config import Configuration
except (ValueError, ImportError) as e:
    raise Exception('You may need to run WebFinder from the root directory (which includes README.md)', e)


import sys, datetime, time, os, requests, socket
from .util.color import Color
from .util.logger import Logger
from .util.process import Process
from .util.pathgetter import PathGetter
from .util.tools import Tools
from .util.getter import Getter


class WebFinder(object):

    def main(self):
        ''' Either performs action based on arguments, or starts attack scanning '''

        self.dependency_check()

        Configuration.initialize()

        self.run()

    def dependency_check(self):
        ''' Check that required programs are installed '''
        required_apps = []
        optional_apps = []
        missing_required = False
        missing_optional = False

        for app in required_apps:
            if not Process.exists(app):
                missing_required = True
                Color.pl('{!} {R}error: required app {O}%s{R} was not found' % app)

        for app in optional_apps:
            if not Process.exists(app):
                missing_optional = True
                Color.pl('{!} {O}warning: recommended app {R}%s{O} was not found' % app)

        if missing_required:
            Color.pl('{!} {R}required app(s) were not found, exiting.{W}')
            sys.exit(-1)

        if missing_optional:
            Color.pl('{!} {O}recommended app(s) were not found')
            Color.pl('{!} {O}WebFinder may not work as expected{W}')

    def run(self):
        '''
            Main program.
            1) Scans for targets, asks user to select targets
            2) Attacks each target
        '''

        get = PathGetter()
        try:
            get.load_iplist()

            now = time.time()
            ts = int(now)
            timestamp = datetime.datetime.fromtimestamp(ts).strftime('%Y-%m-%d %H:%M:%S')
            Logger.pl('     {C}start time {O}%s{W}' % timestamp)
            if get.duplicated > 0:
                Logger.pl('     {C}duplicate {O}%d{C} ip addresses, duplicates ignored {O}%d{C} ips{W}' % (get.len(), get.duplicated))
            else:
                Logger.pl('     {C}duplicate {O}%d{C} ip addresse{W}' % get.len())
            Logger.pl(' ')

            Logger.pl('{+} {W}Connectivity checker{W}')
            ip = None
            try:

                proxy = {}
                if Configuration.proxy != '':
                    proxy = {
                      'http': Configuration.proxy,
                      'https': Configuration.proxy,
                    }

                ip = socket.gethostbyname(Configuration.host)
                r = Getter.general_request(Configuration.base_target.replace('{ip}', ip), proxy=proxy)

                Configuration.main_code = r.status_code
                Configuration.main_length = len(r.text)
                Configuration.main_min_length = float(Configuration.main_length) * 0.93
                Configuration.main_max_length = float(Configuration.main_length) * 1.07

                Logger.pl('{+} {W}Connection test against {C}%s{W} OK! (IP:%s|CODE:%d|SIZE:%d) ' %
                          (Configuration.target, ip, r.status_code, Configuration.main_length))

            except Exception as e:
                if Configuration.proxy != '':
                    Logger.pl('{!} {R}Error connecting to url {O}%s{R} using proxy {O}%s{W}' % (Configuration.target, Configuration.proxy))
                else:
                    Logger.pl('{!} {R}Error connecting to url {O}%s{R} without proxy{W}' % (Configuration.target))
                
                raise e

            if Configuration.proxy_report_to != '':
                try:
                    proxy = {
                      'http': Configuration.proxy_report_to,
                      'https': Configuration.proxy_report_to,
                    }
                    
                    headers = Configuration.user_headers
                    if Configuration.user_agent:
                        headers['User-Agent'] = Configuration.user_agent

                    r = Getter.general_request(Configuration.base_target.replace('{ip}', ip), proxy=proxy)

                    Logger.pl('{+} {W}Connection test against using report to proxy {C}%s{W} OK! (CODE:%d|SIZE:%d) ' %
                              (Configuration.target, r.status_code, len(r.text)))

                except Exception as e:
                    Logger.pl('{!} {R}Error connecting to url {O}%s{R} using {G}report to{R} proxy {O}%s{W}' %
                              (Configuration.target, Configuration.proxy_report_to))
                    raise e

            Logger.pl('{*} {W}Scanning IP address for {C}%s{W} ' % Configuration.target)

            Logger.pl('     ')

        except Exception as e:
            Color.pl("\n{!} {R}Error: {O}%s" % str(e))
            if Configuration.verbose > 1:
                Color.pl('\n{!} {O}Full stack trace below')
                from traceback import format_exc
                Color.p('\n{!}    ')
                err = format_exc().strip()
                err = err.replace('\n', '\n{W}{!} {W}   ')
                err = err.replace('  File', '{W}{D}File')
                err = err.replace('  Exception: ', '{R}Exception: {O}')
                Color.pl(err)
            Configuration.exit_gracefully(1)

        testing = True
        while testing:
            try:
                get.run()
                Logger.pl('     ')

                if os.path.exists("webfinder.restore"):
                    os.remove("webfinder.restore")

                testing = False
            except Exception as e:
                Color.pl("\n{!} {R}Error: {O}%s" % str(e))
                if Configuration.verbose > 0 or True:
                    Color.pl('\n{!} {O}Full stack trace below')
                    from traceback import format_exc
                    Color.p('\n{!}    ')
                    err = format_exc().strip()
                    err = err.replace('\n', '\n{W}{!} {W}   ')
                    err = err.replace('  File', '{W}{D}File')
                    err = err.replace('  Exception: ', '{R}Exception: {O}')
                    Color.pl(err)
                    testing = False
            except KeyboardInterrupt:
                #Color.pl('\n{!} {O}interrupted{W}\n')
                get.pause() # save status and pause the test

                Tools.clear_line()
                print(" ")

                Color.pl('\n{!} {O}interrupted{W}\n')
                testing = False

        now = time.time()
        ts = int(now)
        timestamp = datetime.datetime.fromtimestamp(ts).strftime('%Y-%m-%d %H:%M:%S')
        Logger.pl('{+} {C}End time {O}%s{W}' % timestamp)

        Logger.pl("{+} Finished tests against {C}%s{W}, exiting" % Configuration.target)

        #Configuration.delete_temp()

    def print_banner(self):
        """ Displays ASCII art of the highest caliber.  """
        Color.pl(Configuration.get_banner())


def run():
    requests.packages.urllib3.disable_warnings()

    o = WebFinder()
    o.print_banner()

    try:
        o.main()

    except Exception as e:
        Color.pl('\n{!} {R}Error:{O} %s{W}' % str(e))

        if Configuration.verbose > 0 or True:
            Color.pl('\n{!} {O}Full stack trace below')
            from traceback import format_exc
            Color.p('\n{!}    ')
            err = format_exc().strip()
            err = err.replace('\n', '\n{W}{!} {W}   ')
            err = err.replace('  File', '{W}{D}File')
            err = err.replace('  Exception: ', '{R}Exception: {O}')
            Color.pl(err)

        Color.pl('\n{!} {R}Exiting{W}\n')

    except KeyboardInterrupt:
        Color.pl('\n{!} {O}interrupted, shutting down...{W}')

    Configuration.exit_gracefully(0)


if __name__ == '__main__':
    run()



#!/usr/bin/python3
# -*- coding: UTF-8 -*-
import datetime
from typing import Optional

from requests import Response

from ..util.tools import Tools

import os, subprocess, socket, re, requests, queue, threading, sys, operator, time, json

import os, re, sys, getopt, random
import sys, struct
import base64, string
import socket
import hashlib
from collections import defaultdict
from OpenSSL import SSL
from OpenSSL.crypto import dump_certificate, dump_publickey, FILETYPE_ASN1, FILETYPE_PEM
from urllib.parse import urlparse

from urllib.parse import urlparse

from ..config import Configuration
from ..util.logger import Logger
from ..util.ssl import HostHeaderSSLAdapter


class Getter:

    '''Static variables'''
    path_found = []
    check_himself = False
    dir_not_found = 404
    not_found_lenght = -1
    checked = 0
    total = 0
    ingore_until = ''
    error_count = 0
    deep_links = []


    '''Local non-static variables'''
    q = queue.Queue()
    iplist = []
    base_url = ''
    last = {}
    running=True
    proxy={}
    paused=False

    def __init__(self, ip_list, check_himself = True):
        self.iplist = ip_list
        Getter.check_himself = check_himself
        Getter.checked = 0
        Getter.total = 0
        Getter.ingore_until = ''
        Getter.running=True

        requests.packages.urllib3.disable_warnings()
        requests.packages.urllib3.util.ssl_.DEFAULT_CIPHERS = 'ALL:@SECLEVEL=1'

        Getter.proxy={}
        if Configuration.proxy != '':
            Getter.proxy = {
              'http': Configuration.proxy,
              'https': Configuration.proxy,
            }
        

        pass

    def pause(self):
        Getter.paused=True

    def add_checked(self):
        if Getter.checked < Getter.total:
            Getter.checked += 1

    def stop(self):
        Getter.running=False

    def run(self, base_url):
        Getter.paused = False
        Getter.running=True
        Getter.path_found = []
        Getter.base_url = base_url

        if Getter.base_url.endswith('/'):
            Getter.base_url = Getter.base_url [:-1]

        for i in range(Configuration.tasks):
            self.last[i] = ''
            t = threading.Thread(target=self.worker, kwargs=dict(index=i))
            t.daemon = True
            t.start()

        insert = True
        if self.ingore_until != '':
            insert = False

        with self.q.mutex:
            self.q.queue.clear()

        Getter.total = len(self.iplist)
        for item in self.iplist:
            if Getter.running and item.strip() != '':
                if not insert and item == self.ingore_until:
                    insert = True

                if insert:
                    self.q.put(item)
                else:
                    self.add_checked()

        self.q.join()  # block until all tasks are done

        while Getter.running:
            if len(self.q.queue) > 0:

                if Configuration.verbose > 5:
                    Logger.pl('{?} {G}Queue len: {O}%s{W}' % len(self.q.queue))

                time.sleep(0.3)
            else:
                Getter.running=False

        Tools.clear_line()

        return Getter.path_found

    @staticmethod
    def general_request(url, proxy=None, force_method=None):

        headers = Configuration.user_headers
        if Configuration.user_agent:
            headers['User-Agent'] = Configuration.user_agent

        # Always define host header
        headers['Host'] = Configuration.host
        headers['If-Modified-Since'] = (
                datetime.datetime.utcnow() - datetime.timedelta(hours=43800, minutes=50)).strftime("%c")

        method = Configuration.request_method.upper()
        if force_method is not None:
            method = force_method.upper()

        url_p = urlparse(url)

        s = requests.Session()

        if url_p.scheme.lower() == "https":
            s.mount('https://', HostHeaderSSLAdapter())

        if method == "POST":
            return s.post(url, verify=False, timeout=30, data={}, headers=headers, allow_redirects=True,
                          proxies=(proxy if proxy is not None else Getter.proxy))
        elif method == "PUT":
            return s.put(url, verify=False, timeout=30, data={}, headers=headers, allow_redirects=True,
                         proxies=(proxy if proxy is not None else Getter.proxy))
        elif method == "OPTIONS":
            return s.options(url, verify=False, timeout=30, headers=headers, allow_redirects=True,
                             proxies=(proxy if proxy is not None else Getter.proxy))
        else:
            return s.get(url, verify=False, timeout=30, headers=headers, allow_redirects=True,
                         proxies=(proxy if proxy is not None else Getter.proxy))

    def worker(self, index):
        try:
            while Getter.running:

                while Getter.paused:
                    time.sleep(1)
                    if not Getter.running:
                        return

                item = self.q.get()
                try:
                    ret_ok = self.do_work(item)
                    if ret_ok:
                        self.last[index] = item
                        Getter.error_count = 0
                    else:
                        Getter.error_count += 1
                except KeyboardInterrupt as e:
                    raise e
                finally:
                    self.q.task_done()

        except KeyboardInterrupt as e:
            raise e

    def do_work(self, ip):

        self.add_checked()

        if Configuration.verbose > 4:
            Logger.pl('{?} {G}Starting worker to: {O}%s{W}' % ip)

        uri = Configuration.base_target.replace('{ip}', ip)
        ret_ok = self.get_uri(uri)
        if Configuration.check_both:
            if 'https://' in uri.lower():
                self.get_uri(uri.replace('https://', 'http://'))
            else:
                self.get_uri(uri.replace('http://', 'https://'))

        return ret_ok

    def get_uri(self, url):

        if Getter.paused or not Getter.running:
            return

        ret_ok = False
        if Configuration.verbose > 4:
            Tools.clear_line()
            Logger.pl('{?} {G}Testing [%d/%d]: {O}%s{W}' % (Getter.checked, Getter.total, url))

        if not Configuration.full_log:
            Tools.clear_line()
            print(("Testing [%d/%d]: %s" % (Getter.checked, Getter.total, url)), end='\r', flush=True)
        
        try_cnt = 0
        while try_cnt < 5:
            try:

                r = Getter.general_request(url)
                if r is not None and r.status_code > 0:
                    ret_ok = True

                if Configuration.full_log or Configuration.verbose > 4:
                    self.log_url(url, r.status_code, len(r.text))

                self.check_if_rise(url, r.status_code, len(r.text), r)

                if Configuration.verbose > 5:
                    ht = '\r\n'.join([
                        f"  {k}: {v}"
                        for k, v in r.headers.items()
                        if k is not None and k.strip != ''
                    ])
                    Logger.pl('{*} Response header: \r\n{W}%s{W}\r\n\r\n' % ht)

                try_cnt = 4
            except Exception as e:

                Tools.clear_line()
                if Configuration.verbose > 1:
                    Logger.pl('{*} {O}Error loading %s: %s{W}' % (url, e))
                elif Configuration.verbose > 0:
                    Logger.pl('{*} {O}Error loading %s{W}' % url)
                pass

            if try_cnt >= 3:
                time.sleep(0.2 * (try_cnt+1))
            try_cnt = try_cnt+1

            return ret_ok

    def check_if_rise(self, url, status_code, size, response=None):

        if status_code == Configuration.main_code \
                and Configuration.main_min_length <= size <= Configuration.main_max_length:

            server = Tools.get_host(url)
            pad = " " * (15 - len(server))

            waf = self.get_waf(url, response)
            if waf is not None:
                waf = '{R}* %s{W}' % waf
            else:
                waf = ''

            Tools.clear_line()
            Logger.pl('{W}Found: {O}%s{W} %s (CODE:%d|SIZE:%d) %s' % (
                server, pad, status_code, size, waf))

            if Configuration.proxy_report_to != '':
                try:

                    proxy = {
                        'http': Configuration.proxy_report_to,
                        'https': Configuration.proxy_report_to,
                    }

                    Getter.general_request(url, proxy)

                except Exception as e:
                    pass

    @classmethod
    def log_url(cls, url, status, size):

        Logger.pl('+ %s (CODE:%d|SIZE:%d) ' % (
            url, status, size))

    @classmethod
    def get_waf(cls, url, response) -> Optional[str]:
        ht = ''
        try:
            if isinstance(response, Response):
                ht = '    \r\n'.join([
                    str(f"{k}: {v}").lower()
                    for k, v in response.headers.items()
                    if k is not None and k.strip != ''
                ])
        except:
            return None

        try:
            for ws in Configuration.waf_list_short:
                if ws in ht:
                    waf = next(iter([
                        k
                        for k, v in Configuration.waf_list.items()
                        for n in v
                        if ws in n
                    ]), None)
                    if waf is not None:
                        return waf
        except:
            pass

        # Get by DNS
        try:
            host = Tools.get_host(url)
            ht = '\n'.join(socket.getnameinfo((host, 0), 0))
            for ws in Configuration.waf_list_short:
                if ws in ht:
                    waf = next(iter([
                        k
                        for k, v in Configuration.waf_list.items()
                        for n in v
                        if ws in n
                    ]), None)
                    if waf is not None:
                        return waf
        except:
            pass

        # Get by Standard Cert
        try:
            cert_data = ''
            host = Tools.get_host(url)
            uri = urlparse(url)
            port = 443
            if uri.scheme.lower() == 'https':

                try:
                    if ':' in uri.netloc:
                        port = int(uri.netloc.split(':')[1])
                except:
                    pass

                context = SSL.Context(method=SSL.TLS_CLIENT_METHOD)
                context.check_hostname = False

                conn = SSL.Connection(context, socket=socket.socket(socket.AF_INET, socket.SOCK_STREAM))
                conn.settimeout(5)
                conn.connect((host, port))
                conn.setblocking(1)
                conn.do_handshake()
                conn.set_tlsext_host_name(host.encode())
                for (idx, cert) in enumerate(conn.get_peer_cert_chain()):
                    cert_data += f'{cls.get_509_name_str(cert.get_issuer())}\n'
                    cert_data += f'{cls.get_509_name_str(cert.get_subject())}\n'

                    san = cls.get_certificate_san(cert)
                    cert_data += f'{san}\n'

                conn.close()

                for ws in Configuration.waf_list_short:
                    if ws in cert_data:
                        waf = next(iter([
                            k
                            for k, v in Configuration.waf_list.items()
                            for n in v
                            if ws in n
                        ]), None)
                        if waf is not None:
                            return waf
        except:
            pass

        return None

    @classmethod
    def get_certificate_san(cls, x509cert):
        san = ''
        ext_count = x509cert.get_extension_count()
        for i in range(0, ext_count):
            ext = x509cert.get_extension(i)
            if 'subjectAltName' in str(ext.get_short_name()):
                san = ext.__str__()
        return san.lower()

    @classmethod
    def get_509_name_str(cls, name):
        try:
            return "".join("/{:s}={:s}".format(name.decode(), value.decode()) for name, value in name.get_components())
        except:
            return str(name)


#!/usr/bin/python3
# -*- coding: UTF-8 -*-

import string, random, sys, re
from urllib.parse import urlparse

class Tools:

    def __init__(self):
        pass

    @staticmethod
    def random_generator(size=6, chars=string.ascii_uppercase + string.digits):
        return ''.join(random.choice(chars) for x in range(size))


    @staticmethod
    def clear_line():
        sys.stdout.write("\033[K")  # Clear to the end of line

    @staticmethod     
    def permited_char(s):
        if s.isalpha():
            return True
        elif bool(re.match("^[A-Za-z0-9:]*$", s)):
            return True
        elif s == ".":
            return True
        else:
            return False

    @staticmethod
    def get_host(url):
        rUri = urlparse(url)
        host = rUri.netloc.strip(': ')
        if ':' in host:
            host = host.split(':')[0]
        return host

    @staticmethod
    def get_port(url):
        rUri = urlparse(url)
        port = ''
        if rUri.scheme.lower() == 'https':
            port = 443
        else:
            port = 80
        if ':' in rUri.netloc:
            port = rUri.netloc.split(':')[1]
        return port



#!/usr/bin/python2.7
# -*- coding: utf-8 -*-

import time
import signal
import os

from subprocess import Popen, PIPE

from ..util.color import Color
from ..config import Configuration


class Process(object):
    ''' Represents a running/ran process '''

    @staticmethod
    def devnull():
        ''' Helper method for opening devnull '''
        return open('/dev/null', 'w')

    @staticmethod
    def call(command, cwd=None, shell=False):
        '''
            Calls a command (either string or list of args).
            Returns tuple:
                (stdout, stderr)
        '''
        if type(command) is not str or ' ' in command or shell:
            shell = True
            if Configuration.verbose > 1:
                Color.pe("\n {C}[?] {W} Executing (Shell): {B}%s{W}" % command)
        else:
            shell = False
            if Configuration.verbose > 1:
                Color.pe("\n {C}[?]{W} Executing: {B}%s{W}" % command)

        pid = Popen(command, cwd=cwd, stdout=PIPE, stderr=PIPE, shell=shell)
        pid.wait()
        (stdout, stderr) = pid.communicate()

        # Python 3 compatibility
        if type(stdout) is bytes: stdout = stdout.decode('utf-8')
        if type(stderr) is bytes: stderr = stderr.decode('utf-8')


        if Configuration.verbose > 1 and stdout is not None and stdout.strip() != '':
            Color.pe("{P} [stdout] %s{W}" % '\n [stdout] '.join(stdout.strip().split('\n')))
        if Configuration.verbose > 1 and stderr is not None and stderr.strip() != '':
            Color.pe("{P} [stderr] %s{W}" % '\n [stderr] '.join(stderr.strip().split('\n')))

        return (stdout, stderr)

    @staticmethod
    def exists(program):
        ''' Checks if program is installed on this system '''
        p = Process(['which', program])
        stdout = p.stdout().strip()
        stderr = p.stderr().strip()

        if stdout == '' and stderr == '':
            return False

        return True

    def __init__(self, command, devnull=False, stdout=PIPE, stderr=PIPE, cwd=None, bufsize=0):
        ''' Starts executing command '''

        if type(command) is str:
            # Commands have to be a list
            command = command.split(' ')

        self.command = command

        if Configuration.verbose > 1:
            Color.pe("\n {C}[?] {W} Executing: {B}%s{W}" % ' '.join(command))

        self.out = None
        self.err = None
        if devnull:
            sout = Process.devnull()
            serr = Process.devnull()
        else:
            sout = stdout
            serr = stderr

        self.start_time = time.time()

        self.pid = Popen(command, stdout=sout, stderr=serr, cwd=cwd, bufsize=bufsize)

    def __del__(self):
        '''
            Ran when object is GC'd.
            If process is still running at this point, it should die.
        '''
        if self.pid and self.pid.poll() is None:
            self.interrupt()

    def stdout(self):
        ''' Waits for process to finish, returns stdout output '''
        self.get_output()
        if Configuration.verbose > 1 and self.out is not None and self.out.strip() != '':
            Color.pe("{P} [stdout] %s{W}" % '\n [stdout] '.join(self.out.strip().split('\n')))
        return self.out

    def stderr(self):
        ''' Waits for process to finish, returns stderr output '''
        self.get_output()
        if Configuration.verbose > 1 and self.err is not None and self.err.strip() != '':
            Color.pe("{P} [stderr] %s{W}" % '\n [stderr] '.join(self.err.strip().split('\n')))
        return self.err

    def stdoutln(self):
        return self.pid.stdout.readline()

    def stderrln(self):
        return self.pid.stderr.readline()

    def get_output(self):
        ''' Waits for process to finish, sets stdout & stderr '''
        if self.pid.poll() is None:
            self.pid.wait()
        if self.out is None:
            (self.out, self.err) = self.pid.communicate()

        if type(self.out) is bytes:
            self.out = self.out.decode('utf-8')

        if type(self.err) is bytes:
            self.err = self.err.decode('utf-8')

        return (self.out, self.err)

    def poll(self):
        ''' Returns exit code if process is dead, otherwise "None" '''
        return self.pid.poll()

    def wait(self):
        self.pid.wait()

    def running_time(self):
        ''' Returns number of seconds since process was started '''
        return int(time.time() - self.start_time)

    def interrupt(self, wait_time=2.0):
        '''
            Send interrupt to current process.
            If process fails to exit within `wait_time` seconds, terminates it.
        '''
        try:
            pid = self.pid.pid
            cmd = self.command
            if type(cmd) is list:
                cmd = ' '.join(cmd)

            if Configuration.verbose > 1:
                Color.pe('\n {C}[?] {W} sending interrupt to PID %d (%s)' % (pid, cmd))

            os.kill(pid, signal.SIGINT)

            start_time = time.time()  # Time since Interrupt was sent
            while self.pid.poll() is None:
                # Process is still running
                time.sleep(0.1)
                if time.time() - start_time > wait_time:
                    # We waited too long for process to die, terminate it.
                    if Configuration.verbose > 1:
                        Color.pe('\n {C}[?] {W} Waited > %0.2f seconds for process to die, killing it' % wait_time)
                    os.kill(pid, signal.SIGTERM)
                    self.pid.terminate()
                    break

        except OSError as e:
            if 'No such process' in e.__str__():
                return
            raise e  # process cannot be killed


if __name__ == '__main__':
    p = Process('ls')
    print(p.stdout(), p.stderr())
    p.interrupt()

    # Calling as list of arguments
    (out, err) = Process.call(['ls', '-lah'])
    print(out, err)

    print('\n---------------------\n')

    # Calling as string
    (out, err) = Process.call('ls -l | head -2')
    print(out, err)

    print('"reaver" exists:', Process.exists('reaver'))

    # Test on never-ending process
    p = Process('yes')
    # After program loses reference to instance in 'p', process dies.



#!/usr/bin/python3
# -*- coding: UTF-8 -*-

import os, subprocess, socket, re, requests, errno, sys, time, json, signal, base64, hashlib, random
from pathlib import Path
from urllib.parse import urlparse

from .args import Arguments
from .util.color import Color
from .util.logger import Logger
from .util.database import Database
from .version import __version__

class Configuration(object):
    ''' Stores configuration variables and functions for Turbo Search. '''
    version = '0.0.0'

    initialized = False # Flag indicating config has been initialized
    verbose = 0
    target = ''
    ip_list = ''
    out_file = ''
    full_log = False
    cmd_line =''
    restore = ''
    threads_data = None
    restored_uri=''
    restored_paths=[]
    threads_data={}
    proxy=''
    host=''
    proxy_report_to=''
    request_method='GET'
    user_agent=''
    user_headers={}
    ipaddresses=[]
    skip_current=False
    db = None
    statsdb=False
    main_code=0
    main_length=0
    check_both=False
    base_target = False
    main_min_length = 0
    main_max_length = 0
    waf_list = {}
    waf_list_short = []

    @staticmethod
    def initialize():
        '''
            Sets up default initial configuration values.
            Also sets config values based on command-line arguments.
        '''

        # Only initialize this class once
        if Configuration.initialized:
            return
        Configuration.initialized = True

        Configuration.verbose = 0 # Verbosity level.
        Configuration.print_stack_traces = True

        # Overwrite config values with arguments (if defined)
        Configuration.load_from_arguments()


    @staticmethod
    def load_from_arguments():
        ''' Sets configuration values based on Argument.args object '''
        from .args import Arguments

        config_check = 0

        sys.argv[0] = 'webfinder'

        force_restore = any(['-R' in word for word in sys.argv])
        help = any(['-h' in word for word in sys.argv])

        if help:
            args = Arguments().args
        else:
            if not force_restore and os.path.exists("webfinder.restore"):
                ignore = any(['-I' in word for word in sys.argv])
                if not ignore:
                    Color.pl('{!} {W}Restorefile (you have 10 seconds to abort... (use option -I to skip waiting)) from a previous session found, to prevent overwriting, ./webfinder.restore')
                    time.sleep(10)
                os.remove("webfinder.restore")

            args = {}
            if os.path.exists("webfinder.restore"):
                try:
                    with open("webfinder.restore", 'r') as f:
                        restore_data = json.load(f)
                        Configuration.cmd_line = restore_data["command"]
                        Configuration.threads_data = restore_data["threads"]

                except Exception as e:
                    Color.pl('{!} {R}error: invalid restore file\r\n')
                    raise

                args = Arguments(Configuration.cmd_line).args

            else:
                args = Arguments().args
                for a in sys.argv:
                    if a != "-I":
                        Configuration.cmd_line += "%s " % a

        Color.pl('{+} {W}Startup parameters')

        Logger.pl('     {C}command line:{O} %s{W}' % Configuration.cmd_line)

        if args.target:
            from .util.tools import Tools

            Configuration.target = args.target
            if Configuration.target.endswith('/'):
                Configuration.target = Configuration.target[:-1]

            #rUri = urlparse(Configuration.target)
            #Configuration.host = rUri.netloc
            #if ':' in Configuration.host:
            #    Configuration.host = Configuration.host.split(':')[0]
            Configuration.host = Tools.get_host(Configuration.target)
            Configuration.base_target = Configuration.target.replace(Configuration.host, '{ip}')

        if args.tasks:
            Configuration.tasks = args.tasks

        if args.ip_list:
            Configuration.ip_list = args.ip_list

        if args.verbose:
            Configuration.verbose = args.verbose

        if args.out_file:
            Configuration.out_file = args.out_file

        if args.check_both:
            Configuration.check_both = args.check_both

        if args.tasks:
            Configuration.tasks = args.tasks

        if Configuration.tasks < 1:
            Configuration.tasks = 1

        if Configuration.tasks > 256:
            Configuration.tasks = 256

        if Configuration.target == '':
            config_check = 1

        if Configuration.ip_list == '':
            config_check = 1

        if config_check == 1:
            Configuration.mandatory()

        if args.full_log:
            Configuration.full_log = args.full_log

        if args.proxy:
            Configuration.proxy = args.proxy

        if args.report_to:
            Configuration.proxy_report_to = args.report_to

        if args.request_method.upper() == "POST":
            Configuration.request_method = "POST"
        elif args.request_method.upper() == "PUT":
            Configuration.request_method = "PUT"
        elif args.request_method.upper() == "OPTIONS":
            Configuration.request_method = "OPTIONS"
        else:
            Configuration.request_method = "GET"

        if args.random_agent:
            try:
                
                with open(str(Path(__file__).parent) + "/resources/user_agents.txt", 'r') as f:
                    # file opened for writing. write to it here
                    line = next(f)
                    for num, aline in enumerate(f, 2):
                        if random.randrange(num):
                            continue
                        if aline.strip("\r\n").strip() == '':
                            continue
                        Configuration.user_agent = aline.strip("\r\n").strip()
                    
            except IOError as x:
                if x.errno == errno.EACCES:
                    Color.pl('{!} {R}error: could not open ./resources/user_agents.txt {O}permission denied{R}{W}\r\n')
                    Configuration.exit_gracefully(0)
                elif x.errno == errno.EISDIR:
                    Color.pl('{!} {R}error: could not open ./resources/user_agents.txt {O}it is an directory{R}{W}\r\n')
                    Configuration.exit_gracefully(0)
                else:
                    Color.pl('{!} {R}error: could not open ./resources/user_agents.txt{W}\r\n')
                    Configuration.exit_gracefully(0)

        # get list of WAF
        try:

            with open(str(Path(__file__).parent) + "/resources/waf_headers.json", 'r') as f:
                tmp_waf_list = json.loads(f.read())

                if not isinstance(tmp_waf_list, dict):
                    raise Exception("Invalid dictionary!")

                Configuration.waf_list = {
                    k: [
                        str(n).lower()
                        for n in v
                    ]
                    for k, v in tmp_waf_list.items()
                }
                Configuration.waf_list_short = [
                    str(v).lower()
                    for k, vl in tmp_waf_list.items()
                    for v in vl
                ]

        except IOError as x:
            if x.errno == errno.EACCES:
                Color.pl('{!} {R}error: could not open ./resources/waf_headers.json {O}permission denied{R}{W}\r\n')
                Configuration.exit_gracefully(0)
            elif x.errno == errno.EISDIR:
                Color.pl('{!} {R}error: could not open ./resources/waf_headers.json {O}it is an directory{R}{W}\r\n')
                Configuration.exit_gracefully(0)
            else:
                Color.pl('{!} {R}error: could not open ./resources/waf_headers.json{W}\r\n')
                Configuration.exit_gracefully(0)
        except Exception as e:
            Color.pl('{!} {R}error: could not parse ./resources/waf_headers.json {O}%s{W}\r\n' % str(e))
            Configuration.exit_gracefully(0)

        if args.header != '':
            jData = {}
            try:
                jData=json.loads(args.header)
            except:
                Logger.pl('{!} {R}error: could not convert header value {O}%s{R} from an JSON object {W}\r\n' % (args.header))
                Configuration.exit_gracefully(0)

            Configuration.user_headers = {}
            try:
                for k in jData:
                    if isinstance(k, str):
                        if isinstance(jData[k], str):
                            if k.lower().find("user-agent") != -1:
                                Configuration.user_agent = jData[k]
                            elif k.lower().find("host") != -1:
                                pass
                            elif k.lower().find("connection") != -1:
                                pass
                            elif k.lower().find("accept") != -1:
                                pass
                            elif k.lower().find("accept-encoding") != -1:
                                pass
                            else:
                                Configuration.user_headers[k] = jData[k]
                        else:
                            raise Exception("The value of %s id not an String" % k)
                    else:
                        raise Exception("%s id not an String" % k)
            except Exception as e:
                Logger.pl('{!} {R}error: could parse header data: {R}%s{W}\r\n' % (str(e)))
                Configuration.exit_gracefully(0)

        regex = re.compile(
            r'^(?:http|ftp)s?://'  # http:// or https://
            r'(?:(?:[A-Z0-9](?:[A-Z0-9-]{0,61}[A-Z0-9])?\.)+(?:[A-Z]{2,6}\.?|[A-Z0-9-]{2,}\.?)|'  # domain...
            r'localhost|'  # localhost...
            r'\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})'  # ...or ip
            r'(?::\d+)?'  # optional port
            r'(?:/?|[/?]\S+)$', re.IGNORECASE)

        if re.match(regex, Configuration.target) is None:
            Color.pl('{!} {R}error: invalid target {O}%s{R}{W}\r\n' % Configuration.target)
            Configuration.exit_gracefully(0)

        if Configuration.proxy != '' and re.match(regex, Configuration.proxy) is None:
            Color.pl('{!} {R}error: invalid proxy {O}%s{R}{W}\r\n' % Configuration.proxy)
            Configuration.exit_gracefully(0)

        if Configuration.proxy_report_to != '' and re.match(regex, Configuration.proxy_report_to) is None:
            Color.pl('{!} {R}error: invalid report to proxy {O}%s{R}{W}\r\n' % Configuration.proxy_report_to)
            Configuration.exit_gracefully(0)

        if not os.path.isfile(Configuration.ip_list):
            Color.pl('{!} {R}error: word list file not found {O}%s{R}{W}\r\n' % Configuration.ip_list)
            Configuration.exit_gracefully(0)

        if Configuration.out_file != '':
            try:
                with open(Configuration.out_file, 'a') as f:
                    # file opened for writing. write to it here
                    Logger.out_file = Configuration.out_file
                    f.write(Color.sc(Configuration.get_banner()) + '\n')
                    f.write(Color.sc('{+} {W}Startup parameters') + '\n')
                    pass
            except IOError as x:
                if x.errno == errno.EACCES:
                    Color.pl('{!} {R}error: could not open output file to write {O}permission denied{R}{W}\r\n')
                    Configuration.exit_gracefully(0)
                elif x.errno == errno.EISDIR:
                    Color.pl('{!} {R}error: could not open output file to write {O}it is an directory{R}{W}\r\n')
                    Configuration.exit_gracefully(0)
                else:
                    Color.pl('{!} {R}error: could not open output file to write{W}\r\n')
                    Configuration.exit_gracefully(0)

        try:
            with open(Configuration.ip_list, 'r') as f:
                # file opened for writing. write to it here
                pass
        except IOError as x:
            if x.errno == errno.EACCES:
                Logger.pl('{!} {R}error: could not open word list file {O}permission denied{R}{W}\r\n')
                Configuration.exit_gracefully(0)
            elif x.errno == errno.EISDIR:
                Logger.pl('{!} {R}error: could not open word list file {O}it is an directory{R}{W}\r\n')
                Configuration.exit_gracefully(0)
            else:
                Logger.pl('{!} {R}error: could not open word list file {W}\r\n')
                Configuration.exit_gracefully(0)

        Logger.pl('     {C}target:{O} %s{W}' % Configuration.target)
        Logger.pl('     {C}host:{O} %s{W}' % Configuration.host)

        if Configuration.proxy != '':
            Logger.pl('     {C}Proxy:{O} %s{W}' % Configuration.proxy)

        Logger.pl('     {C}tasks:{O} %s{W}' % Configuration.tasks)

        if args.verbose:
            Logger.pl('     {C}option:{O} verbosity level %d{W}' % Configuration.verbose)

        Logger.pl('     {C}request method: {O}%s{W}' % Configuration.request_method)

        if Configuration.user_agent:
            Logger.pl('     {C}user agent: {O}%s{W}' % Configuration.user_agent)

        Logger.pl('     {C}ip address list:{O} %s{W}' % Configuration.ip_list)

        if Configuration.out_file != '':
            Logger.pl('     {C}output file:{O} %s{W}' % Configuration.out_file)

    @staticmethod
    def get_banner():
            """ Displays ASCII art of the highest caliber.  """

            Configuration.version = str(__version__)

            return '''\

       {G}Web Finder {D}v%s{W}{G} by Helvio Junior{W}
       {W}{D}automated web server finder{W}
       {C}{D}https://github.com/helviojunior/webfinder{W}

    ''' % Configuration.version


    @staticmethod
    def mandatory():
        Color.pl('{!} {R}error: missing a mandatory option ({O}-t and -ip{R}){G}, use -h help{W}\r\n')
        Configuration.exit_gracefully(0)

    @staticmethod
    def exit_gracefully(code=0):
        ''' Deletes temp and exist with the given code '''

        exit(code)


    @staticmethod
    def kill(code=0):
        ''' Deletes temp and exist with the given code '''

        os.kill(os.getpid(),signal.SIGTERM)


    @staticmethod
    def dump():
        ''' (Colorful) string representation of the configuration '''
        from .util.color import Color

        max_len = 20
        for key in Configuration.__dict__.keys():
            max_len = max(max_len, len(key))

        result  = Color.s('{W}%s  Value{W}\n' % 'Configuration Key'.ljust(max_len))
        result += Color.s('{W}%s------------------{W}\n' % ('-' * max_len))

        for (key,val) in sorted(Configuration.__dict__.items()):
            if key.startswith('__') or type(val) == staticmethod or val is None:
                continue
            result += Color.s("{G}%s {W} {C}%s{W}\n" % (key.ljust(max_len),val))
        return result

if __name__ == '__main__':
    Configuration.initialize(False)
    print(Configuration.dump())
